(function (global, factory) {
  if (typeof define === "function" && define.amd) {
    define(['exports', './utils'], factory);
  } else if (typeof exports !== "undefined") {
    factory(exports, require('./utils'));
  } else {
    var mod = {
      exports: {}
    };
    factory(mod.exports, global.utils);
    global.createApiHandler = mod.exports;
  }
})(this, function (exports, _utils) {
  'use strict';

  Object.defineProperty(exports, "__esModule", {
    value: true
  });
  exports.createApiHandler = createApiHandler;
  exports.apiHandler = apiHandler;

  var _apiReducers;

  function _defineProperty(obj, key, value) {
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }

    return obj;
  }

  var _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  function apiKeys(name) {
    return _utils.apiStates.map(function (state) {
      return (0, _utils.upcase)(name + '_' + state);
    });
  }

  // Default reducers (this happens when you don't define a reducer per state)
  // Order is: loading, success, error
  var apiStateHandlers = [function (state) {
    return _extends({}, state, { loading: true });
  }, function (state, action) {
    return _extends({}, state, { loading: false, data: action.payload, errors: [] });
  }, function (state, action) {
    return _extends({}, state, { loading: false, data: [], errors: action.payload });
  }];

  // Mapping of the state to their default handler from above
  // NOTE: this one differs from uncommented version also
  var apiReducers = (_apiReducers = {}, _defineProperty(_apiReducers, _utils.apiStates[0], apiStateHandlers[0]), _defineProperty(_apiReducers, _utils.apiStates[1], apiStateHandlers[1]), _defineProperty(_apiReducers, _utils.apiStates[2], apiStateHandlers[2]), _apiReducers);

  /**
   * Build the state to names handler reducer and
   * the handlers object in one to avoid multiple iterations
   * @constructor
   * @param {string} type string - the type constant
   * @returns {object} object returns with reducers and names
   */
  function handlerReducersForType(type) {
    var apiActionTypes = (0, _utils.getApiTypes)(type);
    var reducers = {};
    var namesToState = {};
    apiActionTypes.forEach(function (key, idx) {
      reducers[apiActionTypes[idx]] = apiReducers[_utils.apiStates[idx]];
      namesToState[(0, _utils.upcase)(_utils.apiStates[idx])] = (0, _utils.toApiKey)(apiActionTypes[idx]);
    });
    return { reducers: reducers, namesToState: namesToState };
  }

  // Decorator for a reducer object property function
  // @param String type - string constant type
  // @param Function handlerFn (optional) - Function defining custom reducers that
  //        accepts a single argument of the api types mapped to their
  //        corresponding constant, i.e.
  //        { loading: 'FETCH_ALL_LOADING',
  //          success: 'FETCH_ALL_SUCCESS',
  //          error: 'FETCH_ALL_ERROR' }
  // Usage:
  //   @apiHandler('FETCH_ALL', (apiStates) => {})
  //   i.e:
  //   @apiHandler(types.SAVE_CROP)
  //   handleSaveCrop: (state, action) => {return new state after doing stuff..})
  function createApiHandler(type, handlerFn) {
    // The decorator function
    // target is the object,
    // name is the function name
    // def is the object property definition
    return function decoratedHandler(target) {
      var _handlerReducersForTy = handlerReducersForType(type);

      var reducers = _handlerReducersForTy.reducers;
      var namesToState = _handlerReducersForTy.namesToState;


      // If there is no handle function or the handle function is not
      // a function, then set it to be a identity function
      if (!handlerFn || typeof handlerFn !== 'function') {
        handlerFn = function handlerFn(d) {
          return d;
        };
      }

      // Custom handlers for the different api states from the second argument
      // If no function is passed, the identify function is used
      var fnHandlerKeys = handlerFn(namesToState) || {};

      // Compute difference between the default hanlders and our almost new target
      // Keys our custom function handler overwrites defaults
      var handlerKeys = Object.keys(fnHandlerKeys);

      // All handler keys, i.e. loading, success, error
      var defaultReducerKeys = Object.keys(reducers);

      // The difference between what we handled and what we did not
      // i.e. allows us to only overwrite 1..n without overwriting other defaults
      var missingHandlerKeys = defaultReducerKeys.filter(function (x) {
        return handlerKeys.indexOf(x) === -1;
      });

      var handlers = Object.assign({}, fnHandlerKeys, {});
      // Set defaults for those undefined by custom function
      missingHandlerKeys.forEach(function (k) {
        return handlers[k] = reducers[k];
      });

      // The handler passed by the decorated function
      var valueHandler = target;

      // ALWAYS set the function to be the defined function
      handlers[namesToState.SUCCESS] = valueHandler;

      // Return the new object property definition
      return handlers;
    };
  }

  // decorator form
  // @param type - string constant
  // @return function - a nwe decorated function Object.defineProperty
  function apiHandler(type, handlerFn) {
    // The decorator function wrapper (design: decoratorFn(fn()))
    // @param target - function to be decorated
    // @param name - the function name
    // @param def - the Object.defineProperty definition
    //
    // Functionally, this works by accepting the object definition type
    // from ES5 syntax, i.e.
    //    Object.defineProperty(this, 'fetchAll', {configurable: false, value: 2})
    // and it manipulates the definition by changing the value to be a function
    // that wraps the different api states, aka LOADING, SUCCESS< ERROR
    return function decoration(target, name, def) {
      var valueHandler = def.initializer(type);
      var handlers = createApiHandler(type, handlerFn)(valueHandler);

      // Dark and murky overwrite object in place -- possibly the grossest thing ever
      Object.assign(target, handlers);

      var newDef = {
        enumerable: true,
        writable: true,
        configurable: true,
        value: handlers
      };

      return newDef;
    };
  }

  exports.default = apiHandler;
});